/* Copyright (C) 2004 Datapark corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/
/* 
**  mod_dpsearch.c -- Apache sample dpsearch module
**  [Autogenerated via ``apxs -n dpsearch -g'']
**
*/ 

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"
#include "ap_config.h"


#include "dps_config.h"
#include "dpsearch.h"
#include "dps_xmalloc.h"

/*#define DEBUG*/


#define MOD_DPSEARCH_VERSION_INFO_STRING "mod_dpsearch/" VERSION

#ifdef APACHE2
#define ap_kill_timeout(a)
#define ap_soft_timeout(a, r)
#endif


module 
#ifndef APACHE2
       MODULE_VAR_EXPORT 
#endif
                         dpsearch_module;

/* The sample content handler */
static int dpsearch_handler(request_rec *r) {
  DPS_AGENT *Agent = ap_get_module_config( r->server->module_config, &dpsearch_module );
  DPS_ENV *Env = Agent->Conf;
  DPS_RESULT	*Res;
  DPS_VARLIST	query_vars, Env_Vars;
  const char *old_tmplt, *tmplt, *env;
  const char *bcharset, *lcharset;
	char		*nav = NULL;
	char		*url = NULL;
	char		*searchwords=NULL;
	char		*storedstr=NULL;
	const char      *ResultContentType;
  size_t        catcolumns = 0;
  ssize_t	page1,page2,npages,ppp=10;
  int		res, page_size, page_number;
  size_t	i, swlen = 0, nav_len, ExcerptSize, ExcerptPadding, storedlen;

  if (r->method_number == M_OPTIONS) {
    r->allowed |= (1 << M_GET);
    return DECLINED;
  }

  if (/*strcmp(r->handler, STATUS_MAGIC_TYPE) && */
      strcmp(r->handler, "dpsearch")) {
    return DECLINED;
  }

  if (r->header_only) return OK;

#ifdef DEBUG

#ifdef APACHE1
  ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
#else
  ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
		"dpsearch_handler: Agent %x  Conf %x", Agent, Env);

#endif


  DpsVarListInit(&query_vars);
  DpsVarListInit(&Env_Vars);
  DpsVarListAddLst(&Env_Vars, &Env->Vars, NULL, "*");
/*  DpsVarListDel(&Env->Vars, "np");
  DpsVarListDel(&Env->Vars, "E");
  DpsVarListDel(&Env->Vars, "CS");
  DpsVarListDel(&Env->Vars, "CP");*/

  ap_soft_timeout("mod_dpsearch", r);

  old_tmplt = (char*)DpsStrdup(DpsVarListFindStr(&Env_Vars, "tmplt", "search.htm"));
  DpsParseQueryString(Agent, &Env_Vars, DPS_NULL2EMPTY(r->args));
  env = getenv("CHARSET_SAVED_QUERY_STRING");
  DpsParseQStringUnescaped(&query_vars, env ? env : DPS_NULL2EMPTY(r->args));

#ifdef DEBUG

#ifdef APACHE1
  ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, 
#else
  ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
	       "dpsearch_handler: %s  ", env ? env : DPS_NULL2EMPTY(r->args) );

#endif

  tmplt = (char*)DpsStrdup(DpsVarListFindStr(&Env_Vars, "tmplt", ""));


#ifdef DEBUG

#ifdef APACHE1
  ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, 
#else
  ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
		"dpsearch_handler: old_tmplt: %s  tmplt: %s", old_tmplt, tmplt );

#endif

  
  if (strcasecmp(old_tmplt, tmplt)) {
    char template_name[PATH_MAX];
    const char *conf_dir;

    DpsVarListFree(&Agent->tmpl);
    
    if (!(conf_dir=getenv("DPS_ETC_DIR"))) conf_dir = DPS_CONF_DIR;
    dps_snprintf(template_name, sizeof(template_name), "%s/%s", conf_dir, tmplt);
    
    if ( DPS_OK != DpsTemplateLoad(Agent->Conf, &Env_Vars, template_name, &Agent->tmpl) ) {
#ifdef APACHE1
      ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, 
#else
      ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
		    "Can't load stemplate '%s'", tmplt);

      DPS_FREE(old_tmplt);
      DPS_FREE(tmplt);
      return 
#ifdef APACHE1
		   SERVER_ERROR;
#else
		   HTTP_INTERNAL_SERVER_ERROR;
#endif
    }
  }
  
  /* Call again to load search Limits if need */
  DpsParseQueryString(Agent, &Env_Vars, DPS_NULL2EMPTY(r->args));
  Agent->Flags = Agent->Conf->Flags;
  Agent->flags = Agent->Conf->flags |= DPS_FLAG_SPELL | DPS_FLAG_UNOCON;

  DpsOpenLog("mod_dpsearch", Agent->Conf, 1);
  DpsSetLogLevel(Agent, DpsVarListFindInt(&Agent->Conf->Vars, "LogLevel", 0)  );

#ifdef DEBUG

#ifdef APACHE1
      ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, 
#else
      ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
		    "SetLogLevel: %d", DpsVarListFindInt(&Agent->Conf->Vars, "LogLevel", 0));

#endif


  /* This is for query tracking */
  DpsVarListReplaceStr(&Env_Vars, "QUERY_STRING", DPS_NULL2EMPTY(r->args));
  DpsVarListReplaceStr(&Env_Vars, "self", r->uri);
  DpsVarListReplaceStr(&Env_Vars, "IP", r->connection->remote_ip);

  bcharset = DpsVarListFindStr(&Env_Vars, "BrowserCharset", "iso-8859-1");
  Env->bcs = DpsGetCharSet(bcharset);
  lcharset = DpsVarListFindStr(&Env_Vars, "LocalCharset", "iso-8859-1");
  Env->lcs = DpsGetCharSet(lcharset);
  
  ppp = DpsVarListFindInt(&Env_Vars, "PagesPerScreen", 10);
  r->content_type =  DpsVarListFindStr(&Env_Vars, "ResultContentType", "text/html");
#ifdef APACHE1
  ap_send_http_header(r);
#endif

  if (!Env->bcs) {
      DPS_FREE(old_tmplt);
      DPS_FREE(tmplt);
#ifdef APACHE1
      ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, 
#else
      ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
		    "Unknown BrowserCharset '%s' in template '%s'", bcharset, tmplt);
      return 
#ifdef APACHE1
		   SERVER_ERROR;
#else
		   HTTP_INTERNAL_SERVER_ERROR;
#endif
  }
  if (!Env->lcs) {
      DPS_FREE(old_tmplt);
      DPS_FREE(tmplt);
#ifdef APACHE1
      ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, 
#else
      ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
		    "Unknown LocalCharset '%s' in template '%s'", bcharset, tmplt);
      return 
#ifdef APACHE1
		   SERVER_ERROR;
#else
		   HTTP_INTERNAL_SERVER_ERROR;
#endif
  }

  /* These parameters taken from "variable section of template"*/
  
  res = DpsVarListFindInt(&Env_Vars, "np", 0) * DpsVarListFindInt(&Env_Vars, "ps", 20);
  DpsVarListReplaceInt(&Env_Vars, "pn", res);
	
	catcolumns = (size_t)atoi(DpsVarListFindStr(&Env_Vars,"CatColumns",""));
	
	if(catcolumns){
		DPS_CATEGORY C;
		
		bzero((void*)&C, sizeof(C));
		strcpy(C.addr,DpsVarListFindStr(&Env_Vars,"cat",""));
		if(DPS_OK == DpsCatAction(Agent, &C, DPS_CAT_ACTION_LIST)){
			size_t n=1, l = 0;
			char *catlist = NULL;
			
			for(i = 0; i < C.ncategories; i++) l += 128 + strlen(C.Category[i].path) + strlen(C.Category[i].name);
			if (l > 0) catlist = (char*)DpsMalloc(l);
			if (catlist != NULL) {
			  sprintf(catlist, "<table>\n");
			  for(i = 0; i < C.ncategories; i++){
				if(n==1){
					sprintf(catlist+strlen(catlist),"<tr>\n");
				}
				sprintf(catlist+strlen(catlist),"<td><a href=\"?cat=%s\">%s</A></td><td width=60>&nbsp;</td>\n",
						C.Category[i].path,
						C.Category[i].name);
				if(n==catcolumns){
					sprintf(catlist+strlen(catlist),"</tr>\n");
					n=1;
				}else{
					n++;
				}
			  }
			  sprintf(catlist + strlen(catlist), "</table>\n");
			  DpsVarListReplaceStr(&Env_Vars, "CS", catlist);
			  DPS_FREE(catlist);
			}
		}else{
			DpsVarListReplaceStr(&Env_Vars,"E",DpsEnvErrMsg(Agent->Conf));
			DpsTemplatePrint(Agent, (DPS_OUTPUTFUNCTION)&ap_rprintf, r, NULL, 0, &Env_Vars, &Agent->tmpl, "top");
			DpsTemplatePrint(Agent, (DPS_OUTPUTFUNCTION)&ap_rprintf, r, NULL, 0, &Env_Vars, &Agent->tmpl, "error");
			goto end;
		}

		DPS_FREE(C.Category);
		bzero((void*)&C, sizeof(C));
		strcpy(C.addr,DpsVarListFindStr(&Env_Vars, "cat", ""));
		if(DPS_OK == DpsCatAction(Agent, &C, DPS_CAT_ACTION_PATH)){
			char *catpath = NULL;
			size_t l = 0;
			
			for(i = 0; i < C.ncategories; i++) l += 32 + strlen(C.Category[i].path) + strlen(C.Category[i].name);
			catpath = (char*)DpsMalloc(l);
			if (catpath != NULL) {
			  catpath[0] = '\0';
			  for(i = 0; i < C.ncategories; i++){
				sprintf(catpath+strlen(catpath),"/<a href=\"?cat=%s\">%s</A>",
					(C.Category[i].path) ? C.Category[i].path : "",
					(C.Category[i].name) ? C.Category[i].name : "");
			  }
			  DpsVarListReplaceStr(&Env_Vars, "CP", catpath);
			  DPS_FREE(catpath);
			}
		}else{
			DpsVarListReplaceStr(&Env_Vars,"E",DpsEnvErrMsg(Agent->Conf));
			DpsTemplatePrint(Agent, (DPS_OUTPUTFUNCTION)&ap_rprintf, r, NULL, 0, &Env_Vars, &Agent->tmpl, "top");
			DpsTemplatePrint(Agent, (DPS_OUTPUTFUNCTION)&ap_rprintf, r, NULL, 0, &Env_Vars, &Agent->tmpl, "error");
			goto end;
		}
		DPS_FREE(C.Category);
	}

	DpsAgentStoredConnect(Agent);
	
	if(NULL==(Res=DpsFind(Agent, &Env_Vars))){
		DpsVarListReplaceStr(&Env_Vars, "E", DpsEnvErrMsg(Agent->Conf));
		DpsTemplatePrint(Agent, (DPS_OUTPUTFUNCTION)&ap_rprintf, r, NULL, 0, &Env_Vars, &Agent->tmpl, "top");
		DpsTemplatePrint(Agent, (DPS_OUTPUTFUNCTION)&ap_rprintf, r, NULL, 0, &Env_Vars, &Agent->tmpl, "error");
		if (Res != NULL) goto freeres;
		goto end;
	}
	
	DpsVarListReplaceInt(&Env_Vars,"first",(int)Res->first);
	DpsVarListReplaceInt(&Env_Vars,"last",(int)Res->last);
	DpsVarListReplaceInt(&Env_Vars,"total",(int)Res->total_found);
	DpsTemplatePrint(Agent, (DPS_OUTPUTFUNCTION)&ap_rprintf, r, NULL, 0, &Env_Vars, &Agent->tmpl, "top");
	
	if(!Res->WWList.nwords){
		DpsTemplatePrint(Agent, (DPS_OUTPUTFUNCTION)&ap_rprintf, r, NULL, 0, &Env_Vars, &Agent->tmpl, "noquery");
		goto freeres;
	}
	
	if(!Res->num_rows){
		DpsTemplatePrint(Agent, (DPS_OUTPUTFUNCTION)&ap_rprintf, r, NULL, 0, &Env_Vars, &Agent->tmpl, "notfound");
		goto freeres;
	}
	
	page_size   = DpsVarListFindInt(&Env_Vars, "ps", 20);
	page_number = DpsVarListFindInt(&Env_Vars, "np", 0);
	
	for (i = 0; i < Res->WWList.nwords; i++) {
		swlen += (8 * Res->WWList.Word[i].len) + 2;
	}
	if ((searchwords = DpsXmalloc(swlen)) != NULL) {
		int z=0;
		for (i = 0; i < Res->WWList.nwords; i++) {
			if (Res->WWList.Word[i].count > 0) {
				sprintf(DPS_STREND(searchwords), (z)?"+%s":"%s", Res->WWList.Word[i].word);
				z++;
			}
		}
	}
	storedstr = DpsRealloc(storedstr, storedlen = (1024 + 10 * swlen) );
	if (storedstr == NULL) {
	  DPS_FREE(old_tmplt);
	  DPS_FREE(tmplt);
#ifdef APACHE1
	  ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, 
#else
          ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
			"Can't realloc storedstr");
	  return 
#ifdef APACHE1
		       SERVER_ERROR;
#else
		       HTTP_INTERNAL_SERVER_ERROR;
#endif
	}
	
	npages=(Res->total_found/(page_size?page_size:20)) + ((Res->total_found % (page_size?page_size:20) != 0 ) ?  1 : 0);
	page1=page_number-ppp/2;
	page2=page_number+ppp/2;
	if(page1<0){
		page2-=page1;
		page1=0;
	}else
	if(page2>npages){
		page1-=(page2-npages);
		page2=npages;
	}
	if(page1<0)page1=page1=0;
	if(page2>npages)page2=npages;
	nav = (char *)DpsRealloc(nav, nav_len = (size_t)(page2 - page1 + 2) * (1024 + 1024)); 
	                                                    /* !!! 1024 - limit for navbar0/navbar1 template size */ 
	if (nav == NULL) {
	  DPS_FREE(old_tmplt);
	  DPS_FREE(tmplt);
#ifdef APACHE1
	  ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, 
#else
          ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
			"Can't realloc nav");
	  return 
#ifdef APACHE1
		       SERVER_ERROR;
#else
		       HTTP_INTERNAL_SERVER_ERROR;
#endif
	}
	nav[0] = '\0';
	
	/* build NL NB NR */
	for(i = (size_t)page1; i < (size_t)page2; i++){
		DpsVarListReplaceInt(&query_vars,"np",(int)i);
		DpsBuildPageURL(&query_vars, &url);
		DpsVarListReplaceStr(&Env_Vars,"NH",url);
		DpsVarListReplaceInt(&Env_Vars,"NP",(int)(i+1));
		DpsTemplatePrint(Agent, NULL, NULL, DPS_STREND(nav), nav_len - (nav - DPS_STREND(nav)), &Env_Vars, &Agent->tmpl,
				 (i == (size_t)page_number)?"navbar0":"navbar1");
	}
	DpsVarListReplaceStr(&Env_Vars,"NB",nav);
	
	DpsVarListReplaceInt(&query_vars,"np",page_number-1);
	DpsBuildPageURL(&query_vars, &url);
	DpsVarListReplaceStr(&Env_Vars,"NH",url);
	
	if(Res->first==1){/* First page */
		DpsTemplatePrint(Agent, NULL, NULL, nav, nav_len, &Env_Vars, &Agent->tmpl, "navleft_nop");
		DpsVarListReplaceStr(&Env_Vars,"NL",nav);
	}else{
		DpsTemplatePrint(Agent, NULL, NULL, nav, nav_len, &Env_Vars, &Agent->tmpl, "navleft");
		DpsVarListReplaceStr(&Env_Vars,"NL",nav);
	}
	
	DpsVarListReplaceInt(&query_vars,"np",page_number+1);
	DpsBuildPageURL(&query_vars, &url);
	DpsVarListReplaceStr(&Env_Vars,"NH",url);
	
	DpsVarListReplaceInt(&query_vars, "np", 0);
	DpsVarListDel(&query_vars, "s");
	DpsBuildPageURL(&query_vars, &url);
	DpsVarListReplaceStr(&Env_Vars, "FirstPage", url);

	if(Res->last>=Res->total_found){/* Last page */
		DpsTemplatePrint(Agent, NULL, NULL, nav, nav_len, &Env_Vars, &Agent->tmpl, "navright_nop");
		DpsVarListReplaceStr(&Env_Vars, "NR", nav);
	}else{
		DpsTemplatePrint(Agent, NULL, NULL, nav, nav_len, &Env_Vars, &Agent->tmpl, "navright");
		DpsVarListReplaceStr(&Env_Vars,"NR",nav);
	}
	
	DpsTemplatePrint(Agent, (DPS_OUTPUTFUNCTION)&ap_rprintf, r, NULL, 0, &Env_Vars, &Agent->tmpl, "restop");
	
	ExcerptSize = (size_t)DpsVarListFindInt(&Env_Vars, "ExcerptSize", 256);
	ExcerptPadding = (size_t)DpsVarListFindInt(&Env_Vars, "ExcerptPadding", 40);
	
	for(i=0;i<Res->num_rows;i++){
		DPS_DOCUMENT	*Doc=&Res->Doc[i];
		DPS_MATCH	*Alias;
		char		*aliastr;
		char		*alcopy;
		
		/* Create "Alias" variable */
		alcopy = DpsRemoveHiLightDup(DpsVarListFindStr(&Doc->Sections,"URL",""));
		if (alcopy != NULL) {
		  if((Alias = DpsMatchListFind(&Agent->Conf->Aliases,alcopy,0,NULL))){
		        aliastr = (char*)DpsMalloc(strlen(Alias->arg) + strlen(alcopy) + 1);
			if (aliastr == NULL) {
			  DPS_FREE(old_tmplt);
			  DPS_FREE(tmplt);
#ifdef APACHE1
			  ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, 
#else
			  ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
					"Can't realloc storedstr");
			  return 
#ifdef APACHE1
				       SERVER_ERROR;
#else
				       HTTP_INTERNAL_SERVER_ERROR;
#endif
			}
			sprintf(aliastr, "%s%s", Alias->arg, alcopy + strlen(Alias->pattern));
		  }else{
			aliastr = (char*)DpsStrdup(alcopy);
		  }
		  DpsVarListReplaceStr(&Doc->Sections,"Alias", aliastr);
		}
		DpsFree(aliastr);
		DpsFree(alcopy);
	}		
	for(i=0;i<Res->num_rows;i++){
		DPS_DOCUMENT	*Doc=&Res->Doc[i];
		DPS_CATEGORY	C;
		char		*clist;
		const char	*u, *dm;
		char		*eu, *edm, *Excerpt;
		size_t		cl, sc, ir, clistsize;
		const char	*al;
		urlid_t		dc_url_id = (urlid_t)DpsVarListFindInt(&Doc->Sections, "ID", 0);
		urlid_t		dc_origin_id = (urlid_t)DpsVarListFindInt(&Doc->Sections, "Origin-ID", 0);
		
		/* Skip clones */
		if(dc_origin_id)continue;
		
		clist = (char*)DpsMalloc(2048); 
		if (clist == NULL) {
		  DPS_FREE(old_tmplt);
		  DPS_FREE(tmplt);
#ifdef APACHE1
		  ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, 
#else
                  ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
				"Can't malloc clist");
		  return 
#ifdef APACHE1
			       SERVER_ERROR;
#else
			       HTTP_INTERNAL_SERVER_ERROR;
#endif
		}
		clist[0] = '\0';
		clistsize = 0;
		for(cl=0;cl<Res->num_rows;cl++){
			DPS_DOCUMENT	*Clone=&Res->Doc[cl];
			urlid_t		cl_origin_id = (urlid_t)DpsVarListFindInt(&Clone->Sections, "Origin-ID", 0);
			
			if ((dc_url_id == cl_origin_id) && cl_origin_id){
				DPS_VARLIST	CloneVars;
				
				DpsVarListInit(&CloneVars);
				DpsVarListAddLst(&CloneVars, &Env_Vars, NULL, "*");
				DpsVarListReplaceLst(&CloneVars, &Doc->Sections, NULL, "*");
				DpsVarListReplaceLst(&CloneVars, &Clone->Sections, NULL, "*");
				clist = (char*)DpsRealloc(clist, (clistsize = strlen(clist)) + 2048);
				if (clist == NULL) {
				  DPS_FREE(old_tmplt);
				  DPS_FREE(tmplt);
#ifdef APACHE1
				  ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, 
#else
				  ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
						"Can't realloc clist");
				  return 
#ifdef APACHE1
					       SERVER_ERROR;
#else
					       HTTP_INTERNAL_SERVER_ERROR;
#endif
				}
				DpsTemplatePrint(Agent, NULL, NULL, clist + clistsize, 2048, &CloneVars, &Agent->tmpl, "clone");
				DpsVarListFree(&CloneVars);
			}
		}
		clistsize += 2048;
		
		DpsVarListReplaceStr(&Env_Vars,"CL",clist);
		DpsVarListReplaceInt(&Env_Vars, "ID", dc_url_id);
		
		DpsVarListReplaceStr(&Env_Vars, "URL", DpsVarListFindStr(&Doc->Sections, "URL", ""));
		

		DpsVarListReplaceStr(&Env_Vars, "title", "[no title]");

		/* Pass all found user-defined sections */
		for (ir = 0; ir < 256; ir++)
		for (sc = 0; sc < Doc->Sections.Root[ir].nvars; sc++) {
			DPS_VAR *S = &Doc->Sections.Root[ir].Var[sc];
			DpsVarListReplaceStr(&Env_Vars, S->name, S->val);
		}
		
		bzero((void*)&C, sizeof(C));
		strcpy(C.addr,DpsVarListFindStr(&Doc->Sections,"Category",""));
		if(catcolumns && !DpsCatAction(Agent, &C, DPS_CAT_ACTION_PATH)){
			char *catpath = NULL;
			size_t c, l = 0;

			for(c = 0; c < C.ncategories; c++) l += 32 + strlen(C.Category[c].path) + strlen(C.Category[c].name);
			catpath = (char*)DpsMalloc(l);
			if (catpath != NULL) {
			  *catpath = '\0';
			  for(c = 0; c < C.ncategories; c++){
				sprintf(catpath+strlen(catpath)," &gt; <A HREF=\"?cat=%s\">%s</A> ",
					C.Category[c].path,
					C.Category[c].name);
			  }
			  DpsVarListReplaceStr(&Env_Vars,"DY",catpath);
			  DPS_FREE(catpath);
			}
		}
		DPS_FREE(C.Category);

		al = DpsVarListFindStrTxt(&Doc->Sections, "URL", "");
		DpsVarListReplaceInt(&Doc->Sections, "URL_ID", DpsStrHash32(al));

		u =  DpsVarListFindStr(&Env_Vars, "URL", "");
		eu = (char*)DpsMalloc(strlen(u)*10 + 10);
		if (eu == NULL) {
		  DPS_FREE(old_tmplt);
		  DPS_FREE(tmplt);
#ifdef APACHE1
		  ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, 
#else
		  ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
				"Can't alloc eu");
		  return 
#ifdef APACHE1
			       SERVER_ERROR;
#else
			       HTTP_INTERNAL_SERVER_ERROR;
#endif
		}
		DpsEscapeURL(eu, u);
		dm = DpsVarListFindStr(&Env_Vars, "Last-Modified", "");
		edm = (char*)DpsMalloc(strlen(dm)*10 + 10);
		if (edm == NULL) {
		  DPS_FREE(old_tmplt);
		  DPS_FREE(tmplt);
#ifdef APACHE1
		  ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, 
#else
		  ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
				"Can't alloc edm");
		  return 
#ifdef APACHE1
			       SERVER_ERROR;
#else
			       HTTP_INTERNAL_SERVER_ERROR;
#endif
		}
		DpsEscapeURL(edm, dm);

		dps_snprintf(storedstr, storedlen, "%s?rec_id=%s&amp;DM=%s&amp;DS=%d&amp;L=%s&amp;CS=%s&amp;DU=%s&amp;CT=%s&amp;q=%s",
			 DpsVarListFindStr(&Env_Vars, "StoredocURL", "/cgi-bin/storedoc.cgi"),
			 DpsVarListFindStr(&Doc->Sections, "URL_ID", 0),
			 edm,
			 DpsVarListFindInt(&Env_Vars, "Content-Length", 0),
			 DpsVarListFindStr(&Env_Vars,"Content-Language",""),
			 DpsVarListFindStr(&Env_Vars,"Charset",""),
			 eu,
			 DpsVarListFindStr(&Doc->Sections,"Content-Type",""),
			 searchwords
			 );
		DpsFree(eu);
		DpsFree(edm);


#ifdef DEBUG

#ifdef APACHE1
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, 
#else
		ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
			      "body: %s", DpsVarListFindStr(&Env_Vars,"body","(none)"));

#endif

		/* temporary skip large files */
		/* Excerpt = (Doc->size < 200000) ? DpsExcerptDoc(Agent, Res, Doc, 256) : NULL;*/
		
		Excerpt = DpsExcerptDoc(Agent, Res, Doc, ExcerptSize, ExcerptPadding);
		
		if ((Excerpt != NULL) && (strlen(Excerpt) > 6)) {
			  char *HlExcerpt = DpsHlConvert(&Res->WWList, Excerpt, Env->lcs, Env->bcs);

#ifdef DEBUG


#ifdef APACHE1
		  ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, 
#else
		  ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
				"Exbody: %s", Excerpt);
#ifdef APACHE1
		  ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, 
#else
		  ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
				"Hlbody: %s", HlExcerpt);

			  DpsVarListReplaceStr(&Env_Vars,"body",HlExcerpt);

#endif

			  
			  DPS_FREE(HlExcerpt);
			}

#ifdef DEBUG

#ifdef APACHE1
		  ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server, 
#else
		  ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, 
#endif
				"Z: %s", DpsVarListFindStr(&Doc->Sections, "Z", "<NULL>"));

#endif


		if (Excerpt != NULL && (DpsVarListFindStr(&Doc->Sections, "Z", NULL) == NULL)) {
			DpsVarListReplaceInt(&Env_Vars,"ST",1);
			DpsVarListReplaceStr(&Env_Vars, "stored_href", storedstr);
		} else {
			DpsVarListReplaceInt(&Env_Vars,"ST",0);
			DpsVarListReplaceStr(&Env_Vars, "stored_href", "");
		}
		DPS_FREE(Excerpt);
		
		if ( (sc = DpsVarListFindInt(&Env_Vars, "site", 0)) == 0) {
		  DpsVarListReplaceInt(&query_vars,"np", 0);
		  DpsVarListReplaceInt(&query_vars, "site", DpsVarListFindInt(&Doc->Sections, "Site_id", 0));
		  DpsBuildPageURL(&query_vars, &url);
		  DpsVarListReplaceStr(&Env_Vars, "sitelimit_href", url);
		}

		if (Res->PerSite) {
		  DpsVarListReplaceUnsigned(&Env_Vars, "PerSite", Res->PerSite[i]);
		}

		DpsTemplatePrint(Agent, (DPS_OUTPUTFUNCTION)&ap_rprintf, r, NULL, 0, &Env_Vars, &Agent->tmpl, "res");
		
		/* Revoke all found user-defined sections */
		for (ir = 0; ir < 256; ir++)
		for(sc = 0; sc < Doc->Sections.Root[ir].nvars; sc++){
			DPS_VAR *S = &Doc->Sections.Root[ir].Var[sc];
			DpsVarListDel(&Env_Vars, S->name);
		}
		DpsVarListDel(&Env_Vars, "body"); /* remoke "body" if it's not in doc's info and made from Excerpt */
		
		DpsFree(clist);
	}
	DpsTemplatePrint(Agent, (DPS_OUTPUTFUNCTION)&ap_rprintf, r, NULL, 0, &Env_Vars, &Agent->tmpl, "resbot");
	DPS_FREE(searchwords);
	DPS_FREE(storedstr);
	
freeres:
	DpsResultFree(Res);
	
end:
	DpsTemplatePrint(Agent, (DPS_OUTPUTFUNCTION)&ap_rprintf, r, NULL, 0, &Env_Vars, &Agent->tmpl, "bottom");
	
	DpsVarListFree(&query_vars);
	DpsVarListFree(&Env_Vars);
	DPS_FREE(url);
	DPS_FREE(nav);

      DPS_FREE(old_tmplt);
      DPS_FREE(tmplt);

  ap_rflush(r);
  ap_kill_timeout(r);

  return OK;
}

#ifdef APACHE1
static void dpsearch_mod_init( server_rec *server, pool *p ) {
#else
static int dpsearch_mod_init( apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s ) {
#endif
	ap_add_version_component(
#ifdef APACHE2
				 p,
#endif
				 MOD_DPSEARCH_VERSION_INFO_STRING);
	DpsInit();        /* Initialize dpsearch library */
	DpsInitMutexes();
#ifdef APACHE2
	return OK;
#endif
}
 
 
static 
#ifdef APACHE1
void 
#else
apr_status_t
#endif
ShutdownDpsearch(void *p) {
  DPS_AGENT *Agent = (DPS_AGENT*)p;
  DPS_ENV *Conf = Agent->Conf;

  DpsAgentFree(Agent);
  DpsEnvFree(Conf);
  DpsDestroyMutexes();

#ifdef APACHE2
  return APR_SUCCESS;
#endif

}

#ifdef APACHE2
static apr_status_t ShutdownEmpty(void *p) {
  return APR_SUCCESS;
}
#endif
 

static void *dpsearch_server_init(
#ifdef APACHE1 
				  pool *p,
#else
				  apr_pool_t *p,
#endif
				  server_rec *s) {
  DPS_ENV *Conf;
  DPS_AGENT *Agent;

  Conf = DpsEnvInit(NULL);
  Agent = DpsAgentInit(NULL, Conf, 0);
#ifdef APACHE2
  apr_pool_cleanup_register(p, Agent, ShutdownEmpty, ShutdownDpsearch);
#endif
  return Agent;
}

#ifdef APACHE1
static void dpsearch_child_exit( server_rec *s, pool *p ) {
  DPS_AGENT *Agent = ap_get_module_config( s->module_config, &dpsearch_module );
  ShutdownDpsearch(Agent);
}
#endif

static const char *cmd_template(cmd_parms *cmd, void *mconfig, const char *template_filename) {
  DPS_AGENT *Agent = ap_get_module_config( cmd->server->module_config, &dpsearch_module );

  DpsVarListReplaceStr(&Agent->Conf->Vars, "tmplt", template_filename);
  if ( DPS_OK != DpsTemplateLoad(Agent->Conf, &Agent->Conf->Vars, template_filename, &Agent->tmpl) ) {
    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 
#ifdef APACHE2
		 0,
#endif
		 cmd->server, "Can't load template '%s' : %s", template_filename, Agent->Conf->errstr);
  }
  Agent->Flags = Agent->Conf->Flags;

  return NULL;
}

static const char *cmd_searchd_conf(cmd_parms *cmd, void *mconfig, const char *word1){
  DPS_AGENT *Agent = ap_get_module_config( cmd->server->module_config, &dpsearch_module );
  uint8 flags = DPS_FLAG_SPELL | DPS_FLAG_UNOCON;
  int verb = -1;
  
  if (DPS_OK != DpsEnvLoad(Agent, word1, flags)) {
    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 
#ifdef APACHE2
		 0,
#endif
		 cmd->server, "Can't load searchd config file '%s'", word1);
  }
/*  DpsUniRegCompileAll(Agent->Conf);*/
  DpsOpenLog("mod_dpsearch", Agent->Conf, 1);
  DpsSetLogLevel(Agent, DPS_LOG_EXTRA);
  Agent->flags = Agent->Conf->flags = flags;
  if (Agent->Conf->Flags.PreloadURLData) {
    DpsLog(Agent, verb, "Preloading url data");
    DpsURLDataPreload(Agent);
  }
  Agent->Flags = Agent->Conf->Flags;
  DpsLog(Agent, verb, "mod_dpsearch started with '%s'", word1);
  DpsLog(Agent, verb, "VarDir: '%s'", DpsVarListFindStr(&Agent->Conf->Vars, "VarDir", DPS_VAR_DIR));
  DpsLog(Agent, verb, "Affixes: %d, Spells: %d, Synonyms: %d", 
	 Agent->Conf->Affixes.naffixes, Agent->Conf->Spells.nspell, Agent->Conf->Synonyms.nsynonyms);

  return NULL;
}


#ifdef APACHE2
static const command_rec dpsearch_cmds[] = {
  AP_INIT_TAKE1("DataparkSearchTemplate", cmd_template, NULL, RSRC_CONF | ACCESS_CONF | OR_ALL,
		"DataparkSearch template file name"),
  AP_INIT_TAKE1("DataparkSearchdConf", cmd_searchd_conf, NULL, RSRC_CONF | OR_NONE,
		"DataparkSearch searchd config file name"),
  {NULL}	
};
#else

static const command_rec dpsearch_cmds[] = {
  { "DataparkSearchTemplate", 
    cmd_template, 
    NULL,
    RSRC_CONF | ACCESS_CONF | OR_ALL, 
    TAKE1, 
    "DataparkSearch template file name" 
  },
  { "DataparkSearchdConf", 
    cmd_searchd_conf, 
    NULL, 
    RSRC_CONF | OR_NONE, 
    TAKE1, 
    "DataparkSearch searchd config file name" 
  },
  {NULL}
};

#endif


#ifdef APACHE2

static void dpsearch_register_hooks(apr_pool_t *p)
{
    ap_hook_handler(dpsearch_handler, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_post_config (dpsearch_mod_init, NULL, NULL, APR_HOOK_MIDDLE);
}

/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA dpsearch_module = {
    STANDARD20_MODULE_STUFF, 
    NULL,                  /* create per-dir    config structures */
    NULL,                  /* merge  per-dir    config structures */
    dpsearch_server_init,                  /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    dpsearch_cmds,                  /* table of config file commands       */
    dpsearch_register_hooks  /* register hooks                      */
};


#else


/* Dispatch list of content handlers */
static const handler_rec dpsearch_handlers[] = { 
    { "dpsearch", dpsearch_handler }, 
    { NULL, NULL }
};


/* Dispatch list for API hooks */
module MODULE_VAR_EXPORT dpsearch_module = {
    STANDARD_MODULE_STUFF, 
    dpsearch_mod_init,                  /* module initializer                  */
    NULL,                  /* create per-dir    config structures */
    NULL,                  /* merge  per-dir    config structures */
    dpsearch_server_init,                  /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    dpsearch_cmds,                  /* table of config file commands       */
    dpsearch_handlers,       /* [#8] MIME-typed-dispatched handlers */
    NULL,                  /* [#1] URI to filename translation    */
    NULL,                  /* [#4] validate user id from request  */
    NULL,                  /* [#5] check if the user is ok _here_ */
    NULL,                  /* [#3] check access by host address   */
    NULL,                  /* [#6] determine MIME type            */
    NULL,                  /* [#7] pre-run fixups                 */
    NULL,                  /* [#9] log a transaction              */
    NULL,                  /* [#2] header parser                  */
    NULL,                  /* child_init                          */
    dpsearch_child_exit,                  /* child_exit                          */
    NULL                   /* [#0] post read-request              */
#ifdef EAPI
   ,NULL,                  /* EAPI: add_module                    */
    NULL,                  /* EAPI: remove_module                 */
    NULL,                  /* EAPI: rewrite_command               */
    NULL                   /* EAPI: new_connection                */
#endif
};


#endif
